@@ -1365,13 +1365,10 @@ static struct near_ndef_mime_payload *parse_mime_type(
 		c_temp->size = record->header->payload_len;
 		memcpy(c_temp->data, ndef_data + offset, c_temp->size);
 	} else {
-		g_free(mime->type);
-		g_free(mime);
 		g_free(c_temp);
-		mime = NULL;
 		c_temp = NULL;
 		*c_data = NULL;
-		return NULL;
+		return mime;
 	}
 
 	*c_data = c_temp;
@@ -1650,7 +1647,7 @@ static struct near_ndef_message *near_ndef_prepare_cfg_message(char *mime_type,
 
 	DBG(" %s", mime_type);
 
-	if (mime_type == NULL || data == NULL || data_len <= 0)
+	if (mime_type == NULL)
 		return NULL;
 
 	msg = ndef_message_alloc_complete(mime_type, data_len, &cdr, cdr_len,
@@ -1659,7 +1656,8 @@ static struct near_ndef_message *near_ndef_prepare_cfg_message(char *mime_type,
 		return NULL;
 
 	/* store data */
-	memcpy(msg->data + msg->offset, data, data_len);
+	if (data)
+		memcpy(msg->data + msg->offset, data, data_len);
 
 	return msg;
 }
@@ -1668,7 +1666,8 @@ static struct near_ndef_message *near_ndef_prepare_cfg_message(char *mime_type,
  * Prepare alternative carrier and configuration records
  * (e.g. bluetooth or wifi or Hc)
  */
-static int near_ndef_prepare_ac_and_cfg_records(enum handover_carrier carrier,
+static int near_ndef_prepare_ac_and_cfg_records(enum record_type type,
+					enum handover_carrier carrier,
 					struct near_ndef_message **ac,
 					struct near_ndef_message **cfg,
 					struct near_ndef_mime_payload *mime,
@@ -1714,6 +1713,42 @@ static int near_ndef_prepare_ac_and_cfg_records(enum handover_carrier carrier,
 		return -EINVAL;
 	}
 
+	/*
+	 * WiFi has to be handled a bit differently when we're trying
+	 * to send the first Hr or Hs (i.e. remote_carrier == NULL).
+	 * There are 2 invalid combinations:
+	 *	a) AP mode and sending an Hr, as we won't be able to
+	 *	   initiate the association.
+	 *	b) STA mode and sending an Hs, as the AP won't be able
+	 *	   initiate the association.
+	 *
+	 * a) is detected when local_carrier != NULL and record_type is
+	 * Handover Request.
+	 * b) is detected when local_carrier == NULL and record_type is
+	 * Handover Select.
+	 * In those 2 cases we return an error.
+	 */
+	if (carrier == NEAR_CARRIER_WIFI && remote_carrier == NULL) {
+		if (local_carrier != NULL &&
+				type == RECORD_TYPE_WKT_HANDOVER_REQUEST) {
+			g_free(local_carrier);
+			return -EINVAL;
+		}
+
+		if (local_carrier == NULL &&
+				type == RECORD_TYPE_WKT_HANDOVER_SELECT)
+			return -EINVAL;
+
+		if (local_carrier == NULL &&
+				type == RECORD_TYPE_WKT_HANDOVER_REQUEST) {
+			*cfg = near_ndef_prepare_cfg_message(mime_type,
+							NULL, 0, cdr, 1);
+			*ac = near_ndef_prepare_ac_message(CPS_UNKNOWN, cdr);
+
+			return 0;
+		}
+	}
+
 	if (local_carrier == NULL) {
 		DBG("Unable to retrieve local carrier %s data", carrier_string);
 		return -ESRCH;
@@ -1871,7 +1906,7 @@ fail:
 	return NULL;
 }
 
-static struct near_ndef_message *near_ndef_prepare_hs_message(
+static struct near_ndef_message *near_ndef_prepare_hs_reply(
 					GSList *remote_mimes,
 					GSList *remote_cfgs)
 {
@@ -1905,6 +1940,7 @@ static struct near_ndef_message *near_ndef_prepare_hs_message(
 			goto fail;
 
 		ret = near_ndef_prepare_ac_and_cfg_records(
+					RECORD_TYPE_WKT_HANDOVER_SELECT,
 					remote_mime->handover.carrier_type,
 					&ac_msg, &cfg_msg,
 					remote_mime, remote_cfg);
@@ -1998,22 +2034,38 @@ static enum handover_carrier string2carrier(char *carrier)
 	return NEAR_CARRIER_UNKNOWN;
 }
 
-static struct near_ndef_message *near_ndef_prepare_hr_message(GSList *carriers)
+static struct near_ndef_message *
+near_ndef_prepare_ho_message(enum record_type type, GSList *carriers)
 {
-	struct near_ndef_message *hr_msg = NULL;
+	struct near_ndef_message *ho_msg = NULL;
 	struct near_ndef_message *cr_msg = NULL;
 	struct near_ndef_message *ac_msg;
 	struct near_ndef_message *cfg_msg;
 	GList *ac_msgs = NULL, *cfg_msgs = NULL, *temp;
 	uint16_t collision;
-	uint8_t hr_length, hr_pl_length;
+	uint8_t ho_length, ho_pl_length;
 	int ret = -EINVAL;
+	char *str_type;
 
 	DBG("");
 
+	switch (type) {
+	case RECORD_TYPE_WKT_HANDOVER_REQUEST:
+		str_type = "Hr";
+		break;
+
+	case RECORD_TYPE_WKT_HANDOVER_SELECT:
+		str_type = "Hs";
+		break;
+
+	default:
+		return NULL;
+	}
+
 	/* Hr message should have at least one carrier */
 	while (carriers) {
 		ret = near_ndef_prepare_ac_and_cfg_records(
+				type,
 				string2carrier(carriers->data),
 				&ac_msg, &cfg_msg, NULL, NULL);
 		if (ret == 0) {
@@ -2029,27 +2081,31 @@ static struct near_ndef_message *near_ndef_prepare_hr_message(GSList *carriers)
 		goto fail;
 	}
 
-	/* Prepare collision resolution record MB=1 ME=0 */
-	collision = GUINT16_TO_BE(g_random_int_range(0, G_MAXUINT16 + 1));
-	cr_msg = near_ndef_prepare_cr_message(collision);
-	if (cr_msg == NULL)
-		goto fail;
-
-	near_ndef_set_mb_me(cr_msg->data, TRUE, FALSE);
-
 	/* Prepare Hr message */
-	hr_pl_length = 1;
-	hr_pl_length += cr_msg->length;
+	ho_pl_length = 1;
+
+	/* Prepare collision resolution record MB=1 ME=0, only for Hr */
+	if (type == RECORD_TYPE_WKT_HANDOVER_REQUEST) {
+		collision = GUINT16_TO_BE(g_random_int_range(0, G_MAXUINT16 + 1));
+		cr_msg = near_ndef_prepare_cr_message(collision);
+		if (cr_msg == NULL)
+			goto fail;
+
+		near_ndef_set_mb_me(cr_msg->data, TRUE, FALSE);
+
+		ho_pl_length += cr_msg->length;
+	}
 
 	/* Alternative carriers are part of handover record payload length */
-	hr_pl_length += ndef_message_list_length(ac_msgs);
+	ho_pl_length += ndef_message_list_length(ac_msgs);
 
-	hr_length = hr_pl_length;
+	ho_length = ho_pl_length;
 	/* Configuration records are part of handover message length */
-	hr_length += ndef_message_list_length(cfg_msgs);
+	ho_length += ndef_message_list_length(cfg_msgs);
 
-	hr_msg = prepare_handover_message_header("Hr", hr_length, hr_pl_length);
-	if (hr_msg == NULL)
+	ho_msg = prepare_handover_message_header(str_type,
+						ho_length, ho_pl_length);
+	if (ho_msg == NULL)
 		goto fail;
 
 	g_list_foreach(ac_msgs, set_mb_me_to_false, NULL);
@@ -2058,44 +2114,57 @@ static struct near_ndef_message *near_ndef_prepare_hr_message(GSList *carriers)
 	ac_msg = temp->data;
 	near_ndef_set_mb_me(ac_msg->data, FALSE, TRUE);
 
+	/*
+	 * Hs record payloads do not have collision recore, the first record
+	 * will be the first ac record.
+	 */
+	if (type == RECORD_TYPE_WKT_HANDOVER_SELECT) {
+		temp = g_list_first(ac_msgs);
+		ac_msg = temp->data;
+		near_ndef_set_mb_me(ac_msg->data, TRUE, FALSE);
+	}
+
 	g_list_foreach(cfg_msgs, set_mb_me_to_false, NULL);
 	temp = g_list_last(cfg_msgs);
 	cfg_msg = temp->data;
 	near_ndef_set_mb_me(cfg_msg->data, FALSE, TRUE);
 
-	/* copy cr */
-	memcpy(hr_msg->data + hr_msg->offset, cr_msg->data, cr_msg->length);
-	hr_msg->offset += cr_msg->length;
+	if (type == RECORD_TYPE_WKT_HANDOVER_REQUEST) {
+		/* copy cr */
+		memcpy(ho_msg->data + ho_msg->offset,
+				cr_msg->data, cr_msg->length);
+		ho_msg->offset += cr_msg->length;
 
-	if (hr_msg->offset > hr_msg->length)
-		goto fail;
+		if (ho_msg->offset > ho_msg->length)
+			goto fail;
+	}
 
 	/* copy acs */
-	copy_ac_records(hr_msg, ac_msgs);
-	if (hr_msg->offset > hr_msg->length)
+	copy_ac_records(ho_msg, ac_msgs);
+	if (ho_msg->offset > ho_msg->length)
 		goto fail;
 
 	/*
 	 * copy cfgs, cfg (associated to the ac) records length
 	 * (bt or wifi) is not part of Hr initial size.
 	 */
-	copy_cfg_records(hr_msg, cfg_msgs);
+	copy_cfg_records(ho_msg, cfg_msgs);
 
-	DBG("Hr message preparation is done");
+	DBG("Handover message preparation is done");
 
 	free_ndef_message(cr_msg);
 	g_list_free_full(ac_msgs, free_ndef_list);
 	g_list_free_full(cfg_msgs, free_ndef_list);
 
-	return hr_msg;
+	return ho_msg;
 
 fail:
-	near_error("handover Hr record preparation failed");
+	near_error("handover record preparation failed");
 
 	g_list_free_full(ac_msgs, free_ndef_list);
 	g_list_free_full(cfg_msgs, free_ndef_list);
 	free_ndef_message(cr_msg);
-	free_ndef_message(hr_msg);
+	free_ndef_message(ho_msg);
 
 	return NULL;
 }
@@ -2345,8 +2414,15 @@ static struct near_ndef_ho_payload *parse_ho_payload(enum record_type rec_type,
 			goto fail;
 		}
 	} else if (reply != NULL) {
+		/* This is a Hs with no cfg and no Ac: No reply and fail */
+		if (rec_type == RECORD_TYPE_WKT_HANDOVER_SELECT &&
+					g_slist_length(acs) == 0) {
+			*reply = NULL;
+			goto fail;
+		}
+
 		/* Prepare Hs, it depends upon Hr message carrier types */
-		*reply = near_ndef_prepare_hs_message(mimes, c_datas);
+		*reply = near_ndef_prepare_hs_reply(mimes, c_datas);
 		if (*reply == NULL) {
 			DBG("error in preparing in HS record");
 			goto fail;
@@ -2577,9 +2653,13 @@ GList *near_ndef_parse_msg(uint8_t *ndef_data, size_t ndef_length,
 						ndef_length, offset,
 						record->header->payload_len,
 						&c_data);
-			if (record->mime == NULL || c_data == NULL)
+			if (record->mime == NULL)
 				goto fail;
 
+			/* No carrier data, move on */
+			if (c_data == NULL)
+				break;
+
 			if (process_mime_type(record->mime, c_data) < 0) {
 				g_free(c_data);
 				c_data = NULL;
@@ -3053,7 +3133,8 @@ static struct near_ndef_message *build_sp_record(DBusMessage *msg)
 						(uint8_t *)(uri + id_len));
 }
 
-static struct near_ndef_message *build_ho_record(DBusMessage *msg)
+static struct near_ndef_message *build_ho_record(enum record_type type,
+							DBusMessage *msg)
 {
 	struct near_ndef_message *ho;
 	GSList *carriers;
@@ -3061,7 +3142,7 @@ static struct near_ndef_message *build_ho_record(DBusMessage *msg)
 	DBG("");
 
 	carriers = get_carrier_field(msg);
-	ho = near_ndef_prepare_hr_message(carriers);
+	ho = near_ndef_prepare_ho_message(type, carriers);
 	g_slist_free_full(carriers, g_free);
 
 	return ho;
@@ -3091,6 +3172,9 @@ static void get_wsc_data(DBusMessageIter iter, char **ssid, char **pass)
 {
 	DBG("");
 
+	*ssid = NULL;
+	*pass = NULL;
+
 	while (dbus_message_iter_get_arg_type(&iter) !=
 					DBUS_TYPE_INVALID) {
 		const char *key;
@@ -3213,9 +3297,38 @@ static struct near_ndef_message *build_mime_record(DBusMessage *msg)
 			dbus_message_iter_get_basic(&var_iter, &mime_str);
 
 			if (g_strcmp0(mime_str, WIFI_WSC_MIME_STRING) == 0) {
+				struct near_ndef_message *mime;
+				struct carrier_data *carrier;
+
 				get_wsc_data(arr_iter, &ssid, &passphrase);
-				return near_ndef_prepare_wsc_record(ssid,
-								passphrase);
+				if (ssid != NULL)
+					return near_ndef_prepare_wsc_record(
+							ssid, passphrase);
+
+				/*
+				 * If we did not get an SSID and optionally
+				 * a passphrase from the DBus message, then
+				 * we try to get one from the WiFi-WSC agent.
+				 */
+				carrier = __near_agent_handover_request_data(
+							HO_AGENT_WIFI, NULL);
+				if (carrier == NULL)
+					return NULL;
+
+				mime = ndef_message_alloc_complete(
+					WIFI_WSC_MIME_STRING, carrier->size,
+					NULL, 0, RECORD_TNF_MIME, TRUE, TRUE);
+				if (mime == NULL) {
+					g_free(carrier);
+					return NULL;
+				}
+
+				memcpy(mime->data + mime->offset,
+					carrier->data, carrier->size);
+
+				g_free(carrier);
+
+				return mime;
 			}
 
 		}
@@ -3270,7 +3383,12 @@ struct near_ndef_message *__ndef_build_from_message(DBusMessage *msg)
 				ndef = build_sp_record(msg);
 				break;
 			} else if (g_strcmp0(value, "Handover") == 0) {
-				ndef = build_ho_record(msg);
+				ndef = build_ho_record(
+					RECORD_TYPE_WKT_HANDOVER_REQUEST, msg);
+				break;
+			} else if (g_strcmp0(value, "StaticHandover") == 0) {
+				ndef = build_ho_record(
+					RECORD_TYPE_WKT_HANDOVER_SELECT, msg);
 				break;
 			} else if (g_strcmp0(value, "MIME") == 0) {
 				ndef = build_mime_record(msg);
